#include "gradientboostingfusion.h"
#include "common/color.h"
#include "common/statistics.h"
#include <limits>
#include <opencv2/core/core.hpp>
#include <opencv2/ml/ml.hpp>
#include <opencv/cv.h>

namespace illumestimators {

GradientBoostingFusion::GradientBoostingFusion() :
	Fusion()
{
}

GradientBoostingFusion::GradientBoostingFusion(const std::vector<IlluminantEstimator*>& estimators) :
	Fusion(estimators)
{

}

GradientBoostingFusion::~GradientBoostingFusion()
{
}

std::string GradientBoostingFusion::name() const
{
	std::stringstream name;

	name << "GradientBoostingFusion";

	return name.str();
}

std::string GradientBoostingFusion::identifier() const
{
	std::stringstream identifier;

	identifier << "GradientBoostingFusion";

	return identifier.str();
}

Illum GradientBoostingFusion::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask) const
{
	std::vector<IlluminantEstimator*> estimators = getEstimators();
	std::vector<cv::Vec3d> estimates;

	for (unsigned int i = 0; i < estimators.size(); i++) {
		Illum estimate = estimators[i]->estimateIlluminant(image, mask);
		estimates.push_back(cv::Vec3d(estimate.b, estimate.g, estimate.r)); // assuming bgr order
	}

	return estimateIlluminant(estimates);
}

Illum GradientBoostingFusion::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const std::vector<cv::Point>& superpixel, const cv::Mat_<unsigned char>& mask) const
{
	return Illum();
}

Illum GradientBoostingFusion::estimateIlluminant(const std::vector<cv::Vec3d>& estimates) const
{
	cv::Vec3d result;

	for (int c = 0; c < 3; c++) {
		cv::Mat_<float> samples(1, estimates.size());

		for (int n = 0; n < samples.cols; n++) {
			samples(0, n) = estimates[n][c];
		}

		CvMat sampleMat = samples;

		result[c] = m_trees[c].predict(&sampleMat);
	}

	Illum est(result[2], result[1], result[0]); // assumes bgr order
	return Color::calculateChromaticities(est);
}

bool GradientBoostingFusion::train(const std::vector<std::string>& imageFiles, const std::vector<std::string>& colorspaces, const std::vector<cv::Vec3d>& illuminants, const std::vector<std::string>& maskFiles)
{
	return true;
}

bool GradientBoostingFusion::train(const std::vector<std::vector<cv::Vec3d> >& estimates, const std::vector<cv::Vec3d>& illuminants)
{
	if (estimates.size() != illuminants.size()) {
		std::cerr << "Dimensions mismatch!" << std::endl;
		return false;
	}

	for (int c = 0; c < 3; c++) {
		cv::Mat_<float> trainData(estimates.size(), estimates[0].size());

		for (int m = 0; m < trainData.rows; m++) {
			for (int n = 0; n < trainData.cols; n++) {
				trainData(m, n) = estimates[m][n][c];
			}
		}

		cv::Mat_<float> responses(illuminants.size(), 1);
		for (int m = 0; m < responses.rows; m++) {
			responses(m, 0) = illuminants[m][c];
		}

//		CvMat trainDataMat = trainData;
//		CvMat responsesMat = responses;

		cv::Mat v = cv::Mat(trainData.cols + 1, 1, CV_8UC1);
		v = cv::Scalar(CV_VAR_ORDERED);
		CvMat varType = v;

		//CvGBTreesParams p(CvGBTrees::SQUARED_LOSS, 50, 1.0, 1.0, 10, true);
		CvGBTreesParams p(CvGBTrees::SQUARED_LOSS, 200, 1.0, 1.0, 20, true);

		cv::Mat empty;

//		Didn't understand this line, left it out
//		m_trees[c].train(trainData, CV_ROW_SAMPLE, responses, cv::Mat_<unsigned char>::zeros(illuminants.size(), 1), cv::Mat_<unsigned char>::zeros(estimates.size(), estimates[0].size()), v, cv::Mat_<unsigned char>::zeros(estimates.size(), estimates[0].size()), p, false);
		m_trees[c].train(trainData, CV_ROW_SAMPLE, responses, empty, empty, v, empty, p, false);
	}

	return true;
}

void GradientBoostingFusion::preprocessImage(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask, cv::Mat_<cv::Vec3d> &outputImage, cv::Mat_<unsigned char> &outputMask) const
{
	outputImage = image.clone();
	if ((image.rows != mask.rows) || (image.cols != mask.rows)) {
		outputMask = cv::Mat_<unsigned char>(outputImage.rows, outputImage.cols, (unsigned char)0);
	} else {
		outputMask = mask.clone();
	}
}

bool GradientBoostingFusion::save(const std::string& filename) const
{
	cv::FileStorage	fs(filename, cv::FileStorage::WRITE);

	if (!fs.isOpened()) {
		return false;
	}

	fs << "name" << "GradientBoostingFusion";

	m_trees[0].write(*fs, "tree_0");
	m_trees[1].write(*fs, "tree_1");
	m_trees[2].write(*fs, "tree_2");

	return true;
}

bool GradientBoostingFusion::load(const std::string& filename)
{
	cv::FileStorage	fs(filename, cv::FileStorage::READ);

	if (!fs.isOpened()) {
		return false;
	}

	std::string name = fs["name"];

	if (name != "GradientBoostingFusion") {
		return false;
	}

	m_trees[0].read(*fs, cvGetFileNodeByName(*fs, NULL, "tree_0"));
	m_trees[1].read(*fs, cvGetFileNodeByName(*fs, NULL, "tree_1"));
	m_trees[2].read(*fs, cvGetFileNodeByName(*fs, NULL, "tree_2"));

	return true;
}

int GradientBoostingFusion::error()
{
	return 0;
}

} // namespace illumestimators
