#include "gamutmapping3destimator.h"
#include "common/color.h"
#include "common/scale.h"
#include "common/derivative.h"
#include "common/mask.h"
#include <lpsolve/lp_lib.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <iostream>
#include "geometry_utilities.h"

namespace illumestimators {

GamutMapping3DEstimator::GamutMapping3DEstimator(int n, double sigma, int verbosity, bool cache) :
	m_n(n),
	m_sigma(sigma),
	verbosity(verbosity),
	m_cache(cache)
{
	std::vector<cv::Vec3d> vertices;
	vertices.push_back(cv::Vec3d(0, 0, 0));
	vertices.push_back(cv::Vec3d(0, 0, 1));
	vertices.push_back(cv::Vec3d(0, 1, 0));
	vertices.push_back(cv::Vec3d(1, 0, 0));
	vertices.push_back(cv::Vec3d(1, 1, 1));
	setCanonicalGamut(vertices);
}

GamutMapping3DEstimator::GamutMapping3DEstimator(const std::vector<cv::Vec3d>& canonical, int n, double sigma, int verbosity, bool cache) :
	m_n(n),
	m_sigma(sigma),
	verbosity(verbosity),
	m_cache(cache)
{
	setCanonicalGamut(canonical);
}

GamutMapping3DEstimator::~GamutMapping3DEstimator()
{

}

std::string GamutMapping3DEstimator::name() const
{
	std::stringstream name;
	name << "GamutMapping3DEstimator(n = " << m_n << ", sigma = " << m_sigma << ")";

	return name.str();
}

std::string GamutMapping3DEstimator::identifier() const
{
	std::stringstream identifier;
	identifier << "gamut3d-" << m_n << "-" << m_sigma;

	return identifier.str();
}

void GamutMapping3DEstimator::setCanonicalGamut(const std::vector<cv::Vec3d>& canonical)
{
	m_canonical.setPoints(canonical);
	m_canonical.addPoint(cv::Vec3d(0, 0, 0));

	m_facets = m_canonical.facets();
	m_vertices = m_canonical.vertices();
}

Illum GamutMapping3DEstimator::calculateMapping(const std::vector<cv::Vec3d>& vertices) const
{
	lprec* lp = make_lp(0, 3);

	if (lp == NULL)
		return Illum();

	set_verbose(lp, IMPORTANT);

	int colNumber[] = { 1, 2, 3 };
	double row[3];

	set_add_rowmode(lp, true);

	for (unsigned int i = 0; i < vertices.size(); i++) {
		for(unsigned int j = 0; j < m_facets.size(); j++) {
			for (int k = 0; k < 3; k++) {
				row[k] = vertices[i][k] * m_facets[j].normal[k];
			}

			if (!add_constraintex(lp, 3, row, colNumber, LE, -m_facets[j].offset)) {
				return Illum();
			}
		}
	}

	set_add_rowmode(lp, false);

	{
		for (int i = 0; i < 3; i++) {
			row[i] = 1;
		}

		if (!set_obj_fnex(lp, 3, row, colNumber)) {
			return Illum();
		}
	}

	set_maxim(lp);

	if (!(solve(lp) == OPTIMAL)) {
		std::cerr << "Failed to solve linear program!" << std::endl;
		return Illum();
	}

	get_variables(lp, row);

	Illum mapping(row[2], row[1], row[0]); // note bgr order

	delete_lp(lp);

	return mapping;
}

void GamutMapping3DEstimator::preprocessImage(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask, cv::Mat_<cv::Vec3d> &outputImage, cv::Mat_<unsigned char> &outputMask) const
{
	outputImage = image.clone();
	if ((image.rows != mask.rows) || (image.cols != mask.cols)) {
		std::cerr << "GamutMapping: Error! image and mask size don't fit together (image: " << image.rows << "x" << image.cols << ", mask: " << mask.rows << "x" << mask.cols << ")" << std::endl;
		outputMask = cv::Mat_<unsigned char>(outputImage.rows, outputImage.cols, (unsigned char)0);
	} else {
		outputMask = mask.clone();
	}
	Mask::maskSaturatedPixels(outputImage, outputMask, 1.0);
	Mask::maskMinPixels(outputImage, outputMask, 4.0 / 255);
	if (m_sigma > 0) {
		if (m_n == 0) {
			cv::GaussianBlur(outputImage, outputImage, cv::Size(0, 0), m_sigma, m_sigma);
		} else if (m_n > 0) {
			outputImage = Derivative::normDerivativeFilter(outputImage, m_n, m_sigma);
		}
	}

}

Illum GamutMapping3DEstimator::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const cv::Mat_<unsigned char>& mask) const
{
	Illum estimate(true);

	cv::Mat_<cv::Vec3d> inputImage;
	cv::Mat_<unsigned char> inputMask;

	preprocessImage(image, mask, inputImage, inputMask);

	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, inputMask);
	pixels.push_back(cv::Vec3d(0, 0, 0));

	if (!vole::GeometryUtilities::spansPlane(pixels)) {
		if (verbosity > 0) {
			std::cout << "GamutMapping3DEstimator::estimateIlluminant: pixels for qhull computation are colinear (number of pixels: "
				<< pixels.size() << "), aborted. " << std::endl;
		}
		return Illum();
	}

	std::vector<cv::Vec3d> vertices = ConvexHull<3>(pixels).vertices();

	Illum result = calculateMapping(vertices);

	for (int i = 0; i < 3; i++) {
		if (result[i] == 0) {
			return Illum();
		}

		estimate[i] = 1 / result[i];
	}

	estimate = Color::calculateChromaticities(estimate);

	return estimate;
}

Illum GamutMapping3DEstimator::estimateIlluminant(const cv::Mat_<cv::Vec3d>& image, const superpixels::Superpixel &superpixel, const cv::Mat_<unsigned char>& mask) const
{
	Illum estimate;
	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(image, superpixel, mask);
	pixels.push_back(cv::Vec3d(0, 0, 0));

	if (!vole::GeometryUtilities::spansPlane(pixels)) {
		if (verbosity > 0) {
			std::cout << "GamutMapping3DEstimator::estimateIlluminant: pixels for qhull computation are colinear (number of pixels: "
				<< pixels.size() << "), aborted. " << std::endl;
		}
		return Illum();
	}

	std::vector<cv::Vec3d> vertices = ConvexHull<3>(pixels).vertices();

	Illum result = calculateMapping(vertices);

	for (int i = 0; i < 3; i++) {
		if (result[i] == 0) {
			return Illum();
		}

		estimate[i] = 1 / result[i];
	}

	estimate = Color::calculateChromaticities(estimate);

	return estimate;
}

std::vector<cv::Vec3d> GamutMapping3DEstimator::calculateCanonicalGamut(const std::string& filename, const std::string& colorspace, const cv::Vec3d& illuminant, const cv::Mat_<unsigned char>& mask)
{
	cv::Mat_<cv::Vec3d> image = cleanlyReadImage(filename, colorspace);
	if (verbosity > 0) std::cout << Statistics::min(Statistics::min(image)) << " " << Statistics::max(Statistics::max(image)) << std::endl;

	cv::Mat_<unsigned char> inputMask;
	cv::Mat_<cv::Vec3d> inputImage;
	preprocessImage(image, mask, inputImage, inputMask);
/*
	Mask::maskSaturatedPixels(inputImage, inputMask, 1.0);

	if (m_sigma > 0) {
		if (m_n == 0) {
			cv::GaussianBlur(inputImage, inputImage, cv::Size(0, 0), m_sigma, m_sigma);
		} else if (m_n > 0) {
			inputImage = Derivative::normDerivativeFilter(inputImage, m_n, m_sigma);
		}
	}
*/
	std::vector<cv::Vec3d> pixels = Mask::unmaskedPixels(inputImage, inputMask);

	for (unsigned int i = 0; i < pixels.size(); i++) {
		for (int c = 0; c < 3; c++) {
			pixels[i][c] /= illuminant[c];
		}
	}

	Scale::scaleByMax(pixels);

	ConvexHull<3> hull(pixels);

	return hull.vertices();
}

bool GamutMapping3DEstimator::train(const std::vector<std::string>& imageFiles, const std::vector<std::string>& colorspaces, const std::vector<cv::Vec3d>& illuminants, const std::vector<std::string>& maskFiles)
{
#if 1
//#if 0
	if ((imageFiles.size() != colorspaces.size()) || (imageFiles.size() != illuminants.size())) {
		return false;
	}

	if (imageFiles.size() < 1) {
		return false;
	}

	m_canonical.clear();

	for (unsigned int i = 0; i < imageFiles.size(); i++) {
		if (m_cache) {
			static std::map<std::string, std::vector<cv::Vec3d> > gamutCache;

			// TODO: Resolve proper path, load/save
			if (gamutCache.find(imageFiles[i]) != gamutCache.end()) {
				m_canonical.addPoints(gamutCache[imageFiles[i]]);
			} else {
				// testread image (FIXME extremely clumsy)
				cv::Mat testRead = cv::imread(imageFiles[i], -1);
				cv::Mat_<unsigned char> mask;
				if (maskFiles.size() > i) {
					mask = cv::imread(maskFiles[i], 0);
				} else {
					 mask = cv::Mat_<unsigned char>(testRead.rows, testRead.cols);
				}
				std::vector<cv::Vec3d> gamut = calculateCanonicalGamut(imageFiles[i], colorspaces[i], illuminants[i], mask);
				gamutCache[imageFiles[i]] = gamut;
				m_canonical.addPoints(gamut);
			}
		} else {
			// testread image (FIXME extremely clumsy)
			cv::Mat testRead = cv::imread(imageFiles[i], -1);
			cv::Mat_<unsigned char> mask;
			if (maskFiles.size() > i) {
				mask = cv::imread(maskFiles[i], 0);
			} else {
				 mask = cv::Mat_<unsigned char>(testRead.rows, testRead.cols);
			}
			std::vector<cv::Vec3d> gamut = calculateCanonicalGamut(imageFiles[i], colorspaces[i], illuminants[i], mask);
			m_canonical.addPoints(gamut);
		}
	}

	m_canonical.addPoint(cv::Vec3d(0, 0, 0));
	m_facets = m_canonical.facets();
	m_vertices = m_canonical.vertices();
#endif
	return true;
}

bool GamutMapping3DEstimator::train(const std::vector<std::vector<cv::Vec3d> >& gamuts)
{
	if (gamuts.size() < 1) {
		return false;
	}

	m_canonical.clear();

	for (unsigned int i = 0; i < gamuts.size(); i++) {
		m_canonical.addPoints(gamuts[i]);
	}

	m_canonical.addPoint(cv::Vec3d(0, 0, 0));
	m_facets = m_canonical.facets();
	m_vertices = m_canonical.vertices();

	return true;
}

bool GamutMapping3DEstimator::save(const std::string& filename) const
{
	cv::FileStorage	fs(filename, cv::FileStorage::WRITE);

	if (!fs.isOpened()) {
		return false;
	}

	fs << "name" << "GamutMapping3DEstimator";

	fs << "canonical" << "[";
	for (unsigned int i = 0; i < m_vertices.size(); i++) {
		fs << m_vertices[i];
	}
	fs << "]";

	fs << "n" << m_n;
	fs << "sigma" << m_sigma;

	return true;
}

bool GamutMapping3DEstimator::load(const std::string& filename)
{
	cv::FileStorage	fs(filename, cv::FileStorage::READ);

	if (!fs.isOpened()) {
		return false;
	}

	std::string name = fs["name"];

	if (name != "GamutMapping3DEstimator") {
		return false;
	}

	std::vector<cv::Vec3d> vertices;
	cv::FileNode canonical = fs["canonical"];
	if (canonical.type() == cv::FileNode::SEQ) {
		for (unsigned int i = 0; i < canonical.size(); i++) {
			CV_Assert(canonical[i].type() == cv::FileNode::SEQ && canonical[i].size() == 3);
			vertices.push_back(cv::Vec3d((double) canonical[i][0], (double) canonical[i][1], (double) canonical[i][2]));
		}
	}
	setCanonicalGamut(vertices);

	m_n = fs["n"];
	m_sigma = fs["sigma"];

	return true;
}

int GamutMapping3DEstimator::error() {
	return 0;
}


} // namespace illumestimators
